IF "%CONDA_BUILD%" == "" (
    call D:\bld\pymongo_1589036757222\work\build_env_setup.bat
)
REM ===== end generated header =====
python setup.py install --single-version-externally-managed --record=record.txt